@@ -2,19 +2,19 @@
  *
  * \author Copyright 2000 Scott Fritzinger
  *
- * \note
+ * \par License
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
- * \note
+ * \par
  * This library is distributed in the hope that it will be useful, 
  * but WITHOUT ANY WARRANTY; without even the implied warranty of 
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details. 
  *
- * \note
+ * \par
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
@@ -51,7 +51,7 @@ struct _CameraWidget {
 	float   value_float;
 
 	/* For Radio and Menu */
-	char    choice [100] [64];
+	char    **choice;
 	int     choice_count;
 
 	/* For Range */
@@ -60,12 +60,15 @@ struct _CameraWidget {
 	float   increment;
 
 	/* Child info */
-	CameraWidget *children [64];
+	CameraWidget **children;
 	int           children_count;
 
 	/* Widget was changed */
 	int     changed;
 
+	/* Widget is read only */
+	int	readonly;
+
 	/* Reference count */
 	int     ref_count;
 
@@ -77,6 +80,8 @@ struct _CameraWidget {
 };
 
 /**
+ * \brief Create a new widget.
+ *
  * The function creates a new #CameraWidget of specified type and with
  * given label.
  *
@@ -90,7 +95,6 @@ int
 gp_widget_new (CameraWidgetType type, const char *label, 
 		   CameraWidget **widget) 
 {
-	int x;
 	static int i = 0;
 
 	CHECK_NULL (label && widget);
@@ -108,21 +112,20 @@ gp_widget_new (CameraWidgetType type, const char *label,
 
         (*widget)->ref_count    	= 1;
 	(*widget)->choice_count 	= 0;
+	(*widget)->choice 		= NULL;
+	(*widget)->readonly 		= 0;
 	(*widget)->id			= i++;
 
         /* Clear all children pointers */
-	memset ((*widget)->children, 0, sizeof((*widget)->children));
+	free ((*widget)->children);
+	(*widget)->children = NULL;
 	(*widget)->children_count = 0;
 
-	/* Clear out the choices */
-	for (x = 0; x < sizeof((*widget)->choice)/sizeof((*widget)->choice[0]); x++)
-		strcpy ((*widget)->choice[x], "");
-
 	return (GP_OK);
 }
 
 /**
- * Frees a #CameraWidget
+ * \brief Frees a CameraWidget
  *
  * @param widget the #CameraWidget to be freed
  * @return a gphoto2 error code.
@@ -131,26 +134,28 @@ gp_widget_new (CameraWidgetType type, const char *label,
 int
 gp_widget_free (CameraWidget *widget)
 {
+	int x;
 	CHECK_NULL (widget);
 
 	/* Free children recursively */
 	if ((widget->type == GP_WIDGET_WINDOW) ||
 	    (widget->type == GP_WIDGET_SECTION)) {
-	    	int x;
-
 	    	for (x = 0; x < gp_widget_count_children (widget); x++)
 			gp_widget_free (widget->children[x]);
+		free (widget->children);
 	}
-	    	
+	for (x = 0; x < widget->choice_count; x++)
+		free (widget->choice[x]);
+	free (widget->choice);
+
         if (widget->value_string)
 		free (widget->value_string);
 	free (widget);
-
 	return (GP_OK);
 }
 
 /**
- * Increments the reference count for the #CameraWidget
+ * \brief Increments the reference count for the CameraWidget
  *
  * @param widget a #CameraWidget you want to ref-count
  * @return a gphoto2 error code.
@@ -167,9 +172,9 @@ gp_widget_ref (CameraWidget *widget)
 }
 
 /**
- * Decrements the reference count for the #CameraWidget
+ * \brief Decrements the reference count for the CameraWidget
  *
- * @param widget a #CameraWidget you want to unref
+ * @param widget a CameraWidget you want to unref
  * @return a gphoto2 error code.
  *
  **/
@@ -187,7 +192,7 @@ gp_widget_unref (CameraWidget *widget)
 }
 
 /**
- * Retrieves the information about the widget
+ * \brief Retrieves the information about the widget
  *
  * @param widget a #CameraWidget
  * @param info
@@ -205,7 +210,7 @@ gp_widget_get_info (CameraWidget *widget, const char **info)
 
 
 /**
- * Sets the information about the widget
+ * \brief Sets the information about the widget
  *
  * @param widget a #CameraWidget
  * @param info Information about above widget
@@ -222,7 +227,7 @@ gp_widget_set_info (CameraWidget *widget, const char *info)
 }
 
 /**
- * Gets the name of the widget
+ * \brief Gets the name of the widget
  *
  * @param widget a #CameraWidget
  * @param name Name of above widget
@@ -239,7 +244,7 @@ gp_widget_get_name (CameraWidget *widget, const char **name)
 }
 
 /**
- * Sets the name of the widget
+ * \brief Sets the name of the widget
  *
  * @param widget a #CameraWidget
  * @param name Name of above widget
@@ -258,7 +263,7 @@ gp_widget_set_name (CameraWidget *widget, const char *name)
 
 
 /**
- * Retrieves the unique id of the #CameraWidget
+ * \brief Retrieves the unique id of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param id 
@@ -274,6 +279,17 @@ gp_widget_get_id (CameraWidget *widget, int *id)
 	return (GP_OK);
 }
 
+/**
+ * \brief Tells that the widget has been changed
+ *
+ * @param widget a #CameraWidget
+ * @param changed a boolean whether we changed or not
+ * @return a gphoto2 error code
+ *
+ * Sets the changed of the CameraWidget depending on 
+ * the changed parameter.
+ *
+ **/
 int
 gp_widget_set_changed (CameraWidget *widget, int changed)
 {
@@ -284,7 +300,45 @@ gp_widget_set_changed (CameraWidget *widget, int changed)
 }
 
 /**
- * Retrieves the type of the #CameraWidget
+ * \brief Tells that the widget is readonly
+ *
+ * @param widget a #CameraWidget
+ * @param changed a boolean whether we are readonly or not
+ * @return a gphoto2 error code
+ *
+ * Sets the readonly of the CameraWidget depending on 
+ * the changed parameter.
+ *
+ * Only useful when called from the camera driver.
+ **/
+int
+gp_widget_set_readonly (CameraWidget *widget, int readonly)
+{
+	CHECK_NULL (widget);
+
+	widget->readonly = readonly;
+	return (GP_OK);
+}
+
+/**
+ * \brief Retrieves the readonly state of the #CameraWidget
+ *
+ * @param widget a #CameraWidget
+ * @param readonly
+ * @return a gphoto2 error code.
+ *
+ **/
+int
+gp_widget_get_readonly (CameraWidget *widget, int *readonly) 
+{
+	CHECK_NULL (widget && readonly);
+
+	*readonly = widget->readonly;
+	return (GP_OK);
+}
+
+/**
+ * \brief Retrieves the type of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param type
@@ -301,7 +355,7 @@ gp_widget_get_type (CameraWidget *widget, CameraWidgetType *type)
 }
 
 /**
- * Retrieves the label of the #CameraWidget
+ * \brief Retrieves the label of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param label
@@ -318,16 +372,16 @@ gp_widget_get_label (CameraWidget *widget, const char **label)
 }
 
 /**
- * Sets the value of the widget
+ * \brief Sets the value of the widget
  *
  * @param widget a #CameraWidget
  * @param value 
  * @return a gphoto2 error code.
  *
  * Please pass
- * (char*) for GP_WIDGET_MENU, GP_WIDGET_TEXT,
+ * (char*) for GP_WIDGET_MENU, GP_WIDGET_TEXT, GP_WIDGET_RADIO,
  * (float) for GP_WIDGET_RANGE,
- * (int)   for GP_WIDGET_DATE, GP_WIDGET_TOGGLE, GP_WIDGET_RADIO, and
+ * (int)   for GP_WIDGET_DATE, GP_WIDGET_TOGGLE, and
  * (CameraWidgetCallback) for GP_WIDGET_BUTTON.
  *
  **/
@@ -374,7 +428,7 @@ gp_widget_set_value (CameraWidget *widget, const void *value)
 }
 
 /**
- * Retrieves the value of the #CameraWidget
+ * \brief Retrieves the value of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param value
@@ -410,7 +464,7 @@ gp_widget_get_value (CameraWidget *widget, void *value)
 }
 
 /**
- * Appends a #CameraWidget to a #CameraWidget
+ * \brief Appends a #CameraWidget to a #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param child the #CameraWidget you would like to append to above
@@ -420,6 +474,7 @@ gp_widget_get_value (CameraWidget *widget, void *value)
 int
 gp_widget_append (CameraWidget *widget, CameraWidget *child) 
 {
+	CameraWidget **newlist;
 	CHECK_NULL (widget && child);
 
 	/* Return if they can't have any children */
@@ -427,6 +482,12 @@ gp_widget_append (CameraWidget *widget, CameraWidget *child)
 	    (widget->type != GP_WIDGET_SECTION))
 		return (GP_ERROR_BAD_PARAMETERS);
 
+	if (widget->children_count)
+		newlist = realloc(widget->children,sizeof(CameraWidget*)*(widget->children_count+1));
+	else
+		newlist = malloc(sizeof(CameraWidget*));
+	if (!newlist) return (GP_ERROR_NO_MEMORY);
+	widget->children = newlist;
 	widget->children[widget->children_count] = child;
 	widget->children_count += 1;
 	child->parent = widget;
@@ -436,7 +497,7 @@ gp_widget_append (CameraWidget *widget, CameraWidget *child)
 }
 
 /**
- * Prepends a #CameraWidget to a #CameraWidget
+ * \brief Prepends a #CameraWidget to a #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param child the #CameraWidget you would like to prepend to above
@@ -447,6 +508,7 @@ int
 gp_widget_prepend (CameraWidget *widget, CameraWidget *child) 
 {
 	int x;
+	CameraWidget**	newlist;
 
 	CHECK_NULL (widget && child);
 
@@ -455,6 +517,13 @@ gp_widget_prepend (CameraWidget *widget, CameraWidget *child)
 	    (widget->type != GP_WIDGET_SECTION))
 		return (GP_ERROR_BAD_PARAMETERS);
 
+	if (widget->children_count)
+		newlist = realloc(widget->children,sizeof(CameraWidget*)*(widget->children_count+1));
+	else
+		newlist = malloc(sizeof(CameraWidget*));
+	if (!newlist) return (GP_ERROR_NO_MEMORY);
+	widget->children = newlist;
+
 	/* Shift down 1 */
 	for (x = widget->children_count; x > 0; x--)
 		widget->children[x] = widget->children[x - 1];
@@ -469,7 +538,7 @@ gp_widget_prepend (CameraWidget *widget, CameraWidget *child)
 }
 
 /**
- * Counts the children of the #CameraWidget
+ * \brief Counts the children of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @return a gphoto2 error code or number of children
@@ -484,7 +553,7 @@ gp_widget_count_children (CameraWidget *widget)
 }
 
 /**
- * Retrieves the child number \c child_number of the parent
+ * \brief Retrieves the child number \c child_number of the parent
  *
  * @param widget a #CameraWidget
  * @param child_number the number of the child
@@ -506,7 +575,7 @@ gp_widget_get_child (CameraWidget *widget, int child_number,
 }
 
 /**
- * Retrieves the child with label \c label of the #CameraWidget
+ * \brief Retrieves the child with label \c label of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param label the label of the child
@@ -543,7 +612,7 @@ gp_widget_get_child_by_label (CameraWidget *widget, const char *label,
 }
 
 /**
- * Retrieves the child with id \c id of the widget
+ * \brief Retrieves the child with id \c id of the widget
  *
  * @param widget a #CameraWidget
  * @param id the id of the child
@@ -579,7 +648,7 @@ gp_widget_get_child_by_id (CameraWidget *widget, int id, CameraWidget **child)
 }
 
 /**
- * Retrieves the child with name \c name of the widget
+ * \brief Retrieves the child with name \c name of the widget
  *
  * @param widget a #CameraWidget
  * @param name the name of the child
@@ -614,6 +683,14 @@ gp_widget_get_child_by_name (CameraWidget *widget, const char *name,
 	return (GP_ERROR_BAD_PARAMETERS);
 }
 
+/**
+ * \brief Retrieves the parent of a CameraWidget
+ *
+ * @param widget a #CameraWidget
+ * @param parent the pointer to the parent to return
+ * @return a gphoto2 error code.
+ *
+ **/
 int
 gp_widget_get_parent (CameraWidget *widget, CameraWidget **parent)
 {
@@ -625,7 +702,7 @@ gp_widget_get_parent (CameraWidget *widget, CameraWidget **parent)
 }
 
 /**
- * Retrieves the root of the #CameraWidget
+ * \brief Retrieves the root of the #CameraWidget
  *
  * @param widget a #CameraWidget
  * @param root
@@ -646,7 +723,7 @@ gp_widget_get_root (CameraWidget *widget, CameraWidget **root)
 }
 
 /**
- * Sets some range parameters of the #CameraWidget
+ * \brief Sets some range parameters of the #CameraWidget
  *
  * @param range a #CameraWidget of type GP_WIDGET_RANGE
  * @param min
@@ -671,7 +748,7 @@ gp_widget_set_range (CameraWidget *range, float min, float max, float increment)
 }
 
 /**
- * Retrieves some range parameters of the #CameraWidget
+ * \brief Retrieves some range parameters of the #CameraWidget
  *
  * @param range a #CameraWidget of type GP_WIDGET_RANGE
  * @param min
@@ -696,7 +773,7 @@ gp_widget_get_range (CameraWidget *range, float *min, float *max,
 }
 
 /**
- * Adds a choice to the #CameraWidget
+ * \brief Adds a choice to the #CameraWidget
  *
  * @param widget a #CameraWidget of type GP_WIDGET_RADIO or GP_WIDGET_MENU
  * @param choice
@@ -706,23 +783,27 @@ gp_widget_get_range (CameraWidget *range, float *min, float *max,
 int
 gp_widget_add_choice (CameraWidget *widget, const char *choice) 
 {
+	char **choices;
 	CHECK_NULL (widget && choice);
 	if ((widget->type != GP_WIDGET_RADIO) &&
 	    (widget->type != GP_WIDGET_MENU))
 		return (GP_ERROR_BAD_PARAMETERS);
 
-	/* static list of choices is full */
-	if (widget->choice_count >= sizeof(widget->choice)/sizeof(widget->choice[0]))
-		return (GP_ERROR);
+	if (widget->choice_count) {
+		choices = realloc (widget->choice, sizeof(char*)*(widget->choice_count+1));
+	} else {
+		choices = malloc (sizeof(char*));
+	}
+	if (!choices) return (GP_ERROR_NO_MEMORY);
 
-	strncpy (widget->choice[widget->choice_count], choice, sizeof(widget->choice[0]));
+	widget->choice = choices;
+	widget->choice[widget->choice_count] = strdup(choice);
 	widget->choice_count += 1;
-
 	return (GP_OK);
 }
 
 /**
- * Counts the choices of the #CameraWidget
+ * \brief Counts the choices of the #CameraWidget
  *
  * @param widget a #CameraWidget of type GP_WIDGET_RADIO or GP_WIDGET_MENU
  * @return a gphoto2 error code or number of choices.
@@ -740,7 +821,7 @@ gp_widget_count_choices (CameraWidget *widget)
 }
 
 /**
- * Retrieves the choice number \c choice_number
+ * \brief Retrieves the choice number \c choice_number
  *
  * @param widget a #CameraWidget of type GP_WIDGET_RADIO or GP_WIDGET_MENU
  * @param choice_number
@@ -765,7 +846,7 @@ gp_widget_get_choice (CameraWidget *widget, int choice_number,
 }
 
 /**
- * Tells if the widget has been changed
+ * \brief Tells if the widget has been changed
  *
  * @param widget a #CameraWidget
  * @return a gphoto2 error code or changed flag.
@@ -786,4 +867,3 @@ gp_widget_changed (CameraWidget *widget)
 
         return (val);
 }
-
